/*
"esp8266 WiFi configuration to eeprom manager for Arduino IDE"

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

 - Wifi setup code is based on the idea of nikant at https://github.com/nikant/espWiFi2eeprom
   many thanks to that !!!!
*/



#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <EEPROM.h>

ESP8266WebServer APserver(80);
const char AP_password[] = "esp8266-1";
String WifiMAC = "";

// web page parts
const char APwebPage1[] PROGMEM = "<!DOCTYPE HTML>\n"
                                  "<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"><title>esp8266 WiFi setup</title>\n"
                                  "<style type=\"text/css\">body {text-align:center;font-family: sans-serif;background-color: #000;color: #fff;font-size:1.2em;} a:link, a:visited, a:hover, a:active {color:#fff;}</style>\n"
                                  "</head>\n"
                                  "<body>\n"
                                  "<h1>esp8266 WiFi setup</h1>\n<br>"
                                  "<table style=\"width:100%;border: 1px solid #fff;\"><tr>"
                                  "<th style=\"text-align:center;width:50%;\"><form action='/APsubmit' method='POST'>SSID : <input type=\"text\" name=\"newssid\" id=\"formnewssid\" value=\"\"><br>"
                                  "PASSWORD : <input type=\"password\" name=\"newpass\" value=\"\" size=\"32\" maxlength=\"64\"><br><input type=\"submit\" value=\"Submit\"></form></th>"
                                  "<th style=\"text-align:left;width:50%;\">";

String APwebPage2 =  "</th></tr></table>\n"
                     "<br><br><form action=\"/\" target=\"_top\"><input type=\"submit\" value=\"home / rescan networks\"></form>\n"
                     "<br><br><form action=\"/reboot\" target=\"_top\"><input type=\"submit\" value=\"restart esp8266\"></form>\n"
                     "<br><br><form action=\"/eraseEEPROM\" target=\"_top\"><input type=\"submit\" value=\"! clear whole EEPROM !\"></form>\n"
                     "<br><br><b>- version:" + VERSION + "-</b>\n"
                     "</body></html>";


String APwebstring = "";   // String to display


// clear first bytes of eeprom
void handle_clearAPeeprom() {
  Serial.println(F("! Clearing WIFI SSID + PASSWORD !"));
  wifi_SSID="";
  wifi_PWD="";
  writeWifi();
  server.send(200, F("text/HTML; charset=utf-8"), wwwCommandHeader + "SSID ERASE DONE! <br>Rebooting ESP into AP Mode!");
  delay(500);
  ESP.restart();
}

// clear ALL eeprom settings
void handle_clearAPalleeprom() {
  Serial.println(F("! Clearing ALL settings !"));
  deleteEEPROM();
  server.send(200, F("text/HTML; charset=utf-8"), wwwCommandHeader + "EEPROM ERASE DONE! <br>Rebooting ESP into AP Mode!");
  delay(2000);
  ESP.restart();
}

// restart esp8266
void handle_APrestart() {
  Serial.println(F("! Restarting in 1 sec! !"));
  server.send(200, F("text/HTML; charset=utf-8"), wwwCommandHeader + "Restarting ESP!" + wwwGoBack);
  delay(2000);
  ESP.restart();
}

// responds to local server / call
void handle_AProot() {
  getAPlist();
  String SServerSend;
  SServerSend = FPSTR(APwebPage1);
  SServerSend += APwebstring + "MAC ADRESS  = " + WifiMAC + "<br>" + APwebPage2;
  APserver.send(200, "text/html", SServerSend);
  delay(100);
}

// when the form with ssid and pass is submited write them to eeprom
void handle_APsubmit() {

  wifi_SSID = APserver.arg("newssid");
  wifi_PWD = APserver.arg("newpass");
  if (wifi_SSID != "") {
    writeWifi();
    Serial.println("SSID : " + wifi_SSID);
    Serial.println("Password : " + wifi_PWD);
    
    APwebstring = F("<h2><b><p>Saved to eeprom... restart to boot into new wifi</b></h2>\n");
    delay(10);
    String SServerSend;
    SServerSend = FPSTR(APwebPage1);
    SServerSend += APwebstring + APwebPage2;
    APserver.send(200, "text/html", SServerSend);
    delay(100);
  }
}

// return the connection type for the AP list
String printConnectionType(int thisType) {
  String con_type = "";
  // read connection type and print out the name:
  switch (thisType) {
    case 255:
      return con_type = "WL_NO_SHIELD";
    case 0:
      return con_type = "WL_IDLE_STATUS";
    case 1:
      return con_type = "WL_NO_SSID_AVAIL";
    case 2:
      return con_type = "WL_SCAN_COMPLETED";
    case 3:
      return con_type = "WL_CONNECTED";
    case 4:
      return con_type = "WL_CONNECT_FAILED";
    case 5:
      return con_type = "WL_CONNECTION_LOST";
    case 6:
      return con_type = "WL_DISCONNECTED";
    default:
      return con_type = "?";
  }
}

// return the connection type for the AP list
String printEncryptionType(int thisType) {
  String enc_type = "";
  // read the encryption type and print out the name:
  switch (thisType) {
    case ENC_TYPE_WEP:
      return enc_type = "WEP";
    case ENC_TYPE_TKIP:
      return enc_type = "WPA";
    case ENC_TYPE_CCMP:
      return enc_type = "WPA2";
    case ENC_TYPE_NONE:
      return enc_type = "None";
    case ENC_TYPE_AUTO:
      return enc_type = "Auto";
    default:
      return enc_type = "?";
  }
}

// test if we are connected
boolean testWiFi() {
  int c = 0;
  Serial.print(F("Waiting for Wifi to connect"));
  // c at 60 with delay 500 is for 30 seconds ;)
  // c at 120 with delay 500 is for 60 seconds ;)
  startBlink(0.25);
  while ( c < 120 ) {
  //  Serial.print(".");
  Serial.print("."); 
    if (WiFi.status() == WL_CONNECTED) {
//    if (wifiMulti.run() == WL_CONNECTED) {
      Serial.println("");
//      Serial.println(printConnectionType(wifiMulti.run()));
      Serial.println(printConnectionType(WiFi.status()));
      return 1;
    }
    delay(500);
    checkButton();
    c++;
  }
  Serial.println("");
//  Serial.println(printConnectionType(wifiMulti.run()));
  Serial.println(printConnectionType(WiFi.status()));
  return 0;
}

// get available AP to connect + HTML list to display
void getAPlist() {
  WiFi.disconnect();
  delay(100);
  int n = WiFi.scanNetworks();
  Serial.println(F("Scan done"));
  if (n == 0) {
    Serial.println(F("No networks found :("));
    APwebstring = F("No networks found :(");
    return;
  }

  // sort by RSSI
  int indices[n];
  for (int i = 0; i < n; i++) {
    indices[i] = i;
  }
  for (int i = 0; i < n; i++) {
    for (int j = i + 1; j < n; j++) {
      if (WiFi.RSSI(indices[j]) > WiFi.RSSI(indices[i])) {
        std::swap(indices[i], indices[j]);
      }
    }
  }

  Serial.println("");
  // HTML Print SSID and RSSI for each network found
  APwebstring = F("<ul>");
  for (int i = 0; i < n; i++)
  {
    APwebstring += F("<li>");
    APwebstring += i + 1;
    APwebstring += F(":&nbsp;&nbsp;<b>");
    //APwebstring += F("<a href=\"#\" target=\"_top\" onClick=\"document.getElementById(\'formnewssid\').value=\'");
    APwebstring += F("<a onClick=\"document.getElementById(\'formnewssid\').value=\'");
    APwebstring += WiFi.SSID(indices[i]);
    APwebstring += F("\'\">");
    APwebstring += WiFi.SSID(indices[i]);
    APwebstring += F("</a>");
    APwebstring += F("</b>&nbsp;&nbsp;&nbsp;(");
    APwebstring += WiFi.RSSI(indices[i]);
    APwebstring += F("&nbsp;dBm)&nbsp;&nbsp;&nbsp;");
    APwebstring += printEncryptionType(WiFi.encryptionType(indices[i]));
    APwebstring += F("</li>");
  }
  APwebstring += F("</ul>");
  delay(100);
  Serial.println("");
  for (int i = 0; i < n; i++)
  {
    Serial.print((i + 1));
    Serial.println(". " + WiFi.SSID(indices[i]) + " " + WiFi.RSSI(indices[i]) + " " + printEncryptionType(WiFi.encryptionType(indices[i])));
  }
  Serial.println("");
  delay(100);
}

// setup a soft AP for the user to connect to esp8266 and give a new ssid and pass
void setupWiFiAP() {
  startBlink(0.1);
  WiFi.mode(WIFI_AP);

  // Do a little work to get a unique-ish name for the soft AP. Append the
  // last two bytes of the MAC (HEX'd):
  uint8_t mac[WL_MAC_ADDR_LENGTH];
  WiFi.softAPmacAddress(mac);
  WifiMAC = WiFi.macAddress();
  String macID = String(mac[WL_MAC_ADDR_LENGTH - 2], HEX) +
                 String(mac[WL_MAC_ADDR_LENGTH - 1], HEX);
  macID.toUpperCase();
  String AP_NameString = F("ESP8266 Config-AP ");
  AP_NameString += macID;

  char AP_NameChar[AP_NameString.length() + 1];
  memset(AP_NameChar, 0, AP_NameString.length() + 1);

  for (int i = 0; i < (int)AP_NameString.length(); i++)
    AP_NameChar[i] = AP_NameString.charAt(i);

  getAPlist();

  APserver.on("/", handle_AProot);
  APserver.on("/APsubmit", handle_APsubmit);
  APserver.on("/reboot", handle_APrestart);
  APserver.on("/erasessid", handle_clearAPeeprom);
  APserver.on("/eraseEEPROM", handle_clearAPalleeprom);

  WiFi.softAP(AP_NameChar, AP_password);

  APserver.begin();

  Serial.print(F("SoftAP IP address: "));
  Serial.println(WiFi.softAPIP());
  Serial.println("AP:" + String(AP_NameChar));
  Serial.println("Pwd:"+ String(AP_password));

  int x = 0;
  //delay = 100 and x = 10 for each 1 sec.
  //with delay 100 and x = 600 AP waits 1 min. before reboot
  //with delay 100 and x = 1800 AP waits 3 min. before reboot
  while (WiFi.status() != WL_CONNECTED) {
    APserver.handleClient();
    delay (100);
    checkButton();
    x++;
    if (x > 1800){ESP.restart();} 
  }
}

// main function called externally
void espNKWiFiconnect(String esid,String epass) {

  WiFi.mode(WIFI_STA);
  delay(10);

  Serial.println("");
  Serial.print("SSID: >");
  Serial.print(esid.c_str());
  Serial.println("<");

  Serial.print("PASS: >");
  Serial.print(epass.c_str());
  Serial.println("<");
  // if ssid not empty try to connect
  if (esid != "") {
    // test esid
    //here we put in a name for the WiFi
    //WiFi.hostname(WifiName);
    WiFi.begin(esid.c_str(), epass.c_str());
    // wifiMulti.addAP(esid.c_str(),epass.c_str());
    delay(10);

    boolean testWiFiAP = testWiFi();

    if (testWiFiAP) {
      return;
    } else if (!testWiFiAP) {
      Serial.println(F("Could not connect to SSID!"));
      if (WiFi.status() == WL_CONNECT_FAILED) {
//      if (wifiMulti.run() == WL_CONNECT_FAILED) {
        // we found the AP but connection failed. maybe pass is wrong.
        // clear eeprom and restart esp so we can try again
        Serial.println(F("Connection failed! Maybe SSID/password are wrong."));
        Serial.println(F("Clearing EEPROM and restarting!"));
        delay(500);
        handle_clearAPeeprom();
        delay(100);
        handle_APrestart();
      } else {
        // could not connect to ssid so starting soft AP
        Serial.println(F("Starting AP at port: 80"));
        setupWiFiAP();
      }
    }
  } else {
    // ssid not found in eeprom so starting soft AP
    Serial.println(F("SSID empty! Starting AP at port: 80"));
    setupWiFiAP();
  }
}


